#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <limits.h>
#include <errno.h>
#include <string.h>
#include <ctype.h>

#define MAX_LENGTH 255
#define NOT_END ( (c = getchar()) != '\0') && (c != '\n')

char rnd_char(int num);
int get_passlen(void);
int want_to_save(void);
void get_path(char *path);
void save_password(char *path, char *password, int len);
void post_message(char *path);

int main(int argc, char **argv)
{

	//passlen - length of the password, i - cycle counter,
	//fd - file descriptor, res - multifunctional variable =)
	int passlen, i, fd, res;

	//array where we'll be seeding
	unsigned short seeding[3];

	//c - reading cycle counter
	char password[MAX_LENGTH+1];

	if (argc > 1) {
		res = sscanf(argv[1], "%i", &passlen);
		if (res == 0) 
			passlen = get_passlen();
			
	}
	
	// /dev/random - random number generator, read the random(4)
	fd = open("/dev/random", O_RDONLY);
	if (fd == -1) {
		printf("Random isn't initialized\n");
		exit(1);
	}

	//fill seeding with random bits from /dev/random
	res = read(fd, seeding, 3);
	if (res == -1) {
		printf("Seeding wasn't extracted\n");
		exit(1);
	}
	
	//collect the seeding - lrand48(3) returns random numbers
	seed48(seeding);
	
	//fill each character password with random character
	for (i = 0; i < passlen; i++) {
		do {
			res = lrand48() % 62;
		} while (res < 0);
		password[i] = rnd_char(res);
	}

	//display the password
	password[passlen] = '\0';
	printf("Password:            \t%s\n", password);

	//close random descriptor.
	close(fd);

	if (argc > 2) {
		char path[PATH_MAX];

		if (strcpy(path, argv[2]) == NULL) {
			get_path(path);
		}
		save_password(path, password, passlen);
		post_message(path);
		
	} else if (want_to_save()) {
		char path[PATH_MAX];

		get_path(path);
		save_password(path, password, passlen);
		post_message(path);
	}

	return 0;	
}

// function returns a character
char rnd_char(int num)
{
	if(num < 10)
		return (char)(48 + num);
	else if (num < 36)
        	return (char)(55 + num);
	else
		return (char)(61 + num);
}

// function gets the length
int get_passlen(void)
{
	int len;
	char c;

	printf("Length is:           \t");
	scanf("%i", &len);
	while (NOT_END) {}

	if (len > (MAX_LENGTH + 1) ) {
		printf("You entered too big length\n");
		exit(1);
	}

	return len;
}

// function asks user about saving the password
// returns 1 on positive answer and 0 on negative
int want_to_save(void)
{
	char confirm, c;

	do {
		printf("Save? (y/n)           \t");
		confirm = getchar();

		while (NOT_END) {}

		tolower(confirm);
	} while ( (confirm != 'y') && (confirm != 'n') );

	if (confirm == 'y')
		return 1;
	else
		return 0;
}

// function gets path to file where user wants to save the password
void get_path(char *path)
{
	int len;
	char *res;

	printf("Enter the full path:  \t");
	res = fgets(path, PATH_MAX, stdin);

	if (res == NULL) {
		printf("Password wasn'n saved\n");
		exit(1);
	}

	len = strlen(path);
	
	if (path[len-1] != '\n') {
		printf("Path is too long\n");
		exit(1);
	}

	path[len-1] = '\0';
}

// function saving the password: if file
// where user wants to save the password
// doesn't exist it will be created else
// pasword will be written to the end of
// specified file.
void save_password(char *path, char *password, int len)
{
	int fd, res;

	//try to open specified file
	fd = open(path, O_WRONLY | O_CREAT | O_APPEND, S_IRUSR);
	if (fd == -1) {
		printf("%s\n", strerror(errno));
		exit(1);
	}

	//try to write the password into file
	res = write(fd, password, len);
	if (res == -1) {
		printf("%s\n", strerror(errno));
		exit(1);
	}

	close(fd);
}

// function writes the message with the full
// file path where the password was saved in
void post_message(char *path)
{

	printf("Password was saved in:\t");

	if (path[0] == '.') {
		char cdir[PATH_MAX];
		int len, i;

		getcwd(cdir, PATH_MAX);
		printf("%s", cdir);

		len = strlen(path);
			
		for (i = 0; i < (len-1); i++) {
			path[i] = path[i+1];
		}

		path[len-1] = '\0';
	}

	printf("%s\n", path);

}
